/* onode.c
   -------

   Code for manipulating FCFS O-Nodes and their forks.
   Maybe the Fork stuff should be in a seperate file,
   you're welcome to patch :)

   $Id: onode.c,v 1.15 2003/10/20 07:18:11 stewart Exp $

   (C)2003 Stewart Smith
   Distributed under the GNU Public License

   Debugged with Beer(TM)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <time.h>

#include "testkit/types.h"
#include "super_block.h"
#include "disk.h"
#include "onode.h"
#include "onode_index.h"
#include "space_bitmap.h"
#include "onode_versioned.h"

struct fcfs_onode1 *onode1_new(struct fcfs_disk *disk)
{
  struct fcfs_onode1 *node;

  node = (struct fcfs_onode1*)malloc(sizeof(struct fcfs_onode1));
  if(!node)
    {
      fprintf(stderr,"cannot allocate onode1, exiting\n");
      abort();
    }

  node->magic1 = FCFS_ONODE_MAGIC1;
  node->onode_num = 0;
  node->onode_revision = 0;
  set_bit(FCFS_OFLAG_ForkLeaf,&(node->flags));
  node->use_count = 0;
  node->onode_size = sizeof(struct fcfs_onode1);
  node->forks.leaf.nr = 0;

  return node;
}

struct fcfs_onode1 *onode1_free(struct fcfs_onode1 *onode)
{
  free(onode);
  return NULL;
}

/* 
   FIXME: the allow_internal is nasty hack, should be removed and made elegant 
*/
int onode1_fork_new(struct fcfs_disk *disk, struct fcfs_block_run *onode_br, u64 fork_type, u64 content_length, void* content,int allow_internal)
{
  struct fcfs_disk_block *block;
  struct fcfs_onode1 *onode;

  block = disk_getblock(disk,BR_SECTOR_T(disk,onode_br));
  onode = (struct fcfs_onode1*)block->data;

  onode->forks.leaf.nr++;	/* should be atomic */
  onode->forks.leaf.fork[onode->forks.leaf.nr-1].fork_type = fork_type;
  onode->forks.leaf.fork[onode->forks.leaf.nr-1].content_length = content_length;
  if(content_length <= FCFS_FORK_SMALL_DATA_SIZE && content_length > 0 && allow_internal)
    {				/* Let's make it a small_data node */
      onode->forks.leaf.fork[onode->forks.leaf.nr-1].fork_flags = FCFS_FORK_InForkData;
      if(content!=NULL)
	memcpy(onode->forks.leaf.fork[onode->forks.leaf.nr-1].data.small_data, content,content_length);
      else
	memset(onode->forks.leaf.fork[onode->forks.leaf.nr-1].data.small_data, 0,FCFS_FORK_SMALL_DATA_SIZE);
    }
  else				/* We're  going to need more space! */
    {
      onode->forks.leaf.fork[onode->forks.leaf.nr-1].fork_flags = 0;
      onode->forks.leaf.fork[onode->forks.leaf.nr-1].data.space.leaf.nr = 0;
      
      if(content_length > FCFS_FORK_SMALL_DATA_SIZE && content!=NULL)
	fprintf(stderr,"**FIXME** Need larger objects\n");
      else
	{
#ifdef DEBUG_ONODE
	  fprintf(stderr,"PREALLOCATING ONODE FORK SPACE\n\n");
#endif
	onode1_fork_grow(disk,onode_br,onode->forks.leaf.nr-1,content_length/disk->bsize + ((content_length%disk->bsize)?1:0));}
    }

  disk_writeblock(block);
  disk_freeblock(block);

  return onode->forks.leaf.nr-1;
}

struct fcfs_disk_block *onode1_fork_getblock(struct fcfs_disk *disk, struct fcfs_onode1 *onode,int forknr,u64 blocknr)
{
  int i;
  u64 cur_blocknr;
  struct fcfs_disk_block *block;

  block = NULL;

  if(forknr > onode->forks.leaf.nr)
    {
      fprintf(stderr,"Invalid Fork Number for onode\n");
      abort();
    }

  //  fprintf(stderr,"\n\n\t\t\tGETTING blocknr %llu from leaf.nr %d\n\n",blocknr,onode->forks.leaf.fork[forknr].data.space.leaf.nr);

  cur_blocknr = 0;
  for(i=0;i<onode->forks.leaf.fork[forknr].data.space.leaf.nr;i++)
    {
      if(cur_blocknr+onode->forks.leaf.fork[forknr].data.space.leaf.br[i].len
	 >= blocknr)
	{			/* blocknr in current br */
	  block = disk_getblock(disk,
		    BR_SECTOR_T(disk,
				&(onode->forks.leaf.fork[forknr].data.space.leaf.br[i]))
				+(blocknr-cur_blocknr));
	  //	  fprintf(stderr,"GOT BLOCK #%llu\n",block->blocknr);
	  
	  break;
	}
      else
	{
	  cur_blocknr+=onode->forks.leaf.fork[forknr].data.space.leaf.br[i].len;
	}
    }

  return block;
}

u64 onode1_fork_length(struct fcfs_onode1 *onode,int forknr)
{
  return (onode->forks.leaf.fork[forknr].content_length);
}

int onode1_fork_write_versioned(struct fcfs_disk *disk,struct fcfs_onode1 *onode,  struct fcfs_block_run *onode_br,int forknr,u64 pos,u64 content_length,void* content)
{
  void* olddata;
  struct fcfs_onode_fork_rev rev;

  rev.magic1 = FCFS_ONODE_FORK_REV_MAGIC1;
  rev.time = time(NULL);
  rev.revision = onode->onode_revision;	/* FIXME */
  rev.consistent_revision = onode->onode_revision; /* FIXME */
  rev.operation = FCFS_ONODE_FORK_REV_REPLACE;
  rev.rev_offset = onode1_fork_length(onode,forknr+1);
  rev.rev_length = content_length;
  rev.offset = pos;
  memset(rev.space,0,16);

  olddata = malloc(content_length);
  if(!olddata)
    {fprintf(stderr,"No memory for updating content\n");abort();}
  onode1_fork_read(disk,onode,forknr,pos,content_length,olddata);
  onode1_fork_write(disk,onode_br,forknr+1,onode1_fork_length(onode,forknr+1),content_length,olddata);
  onode1_fork_write(disk,onode_br,forknr+2,onode1_fork_length(onode,forknr+2),sizeof(struct fcfs_onode_fork_rev),&rev);
  onode1_fork_write(disk,onode_br,forknr,pos,content_length,content);
  return 1;
}

int onode1_fork_write(struct fcfs_disk *disk, struct fcfs_block_run *onode_br,int forknr, u64 pos, u64 content_length, void* content)
{
  struct fcfs_disk_block *onode_block;
  struct fcfs_disk_block *block;
  struct fcfs_onode1 *onode;
  u64 write_length;
  u64 done_length;
  
  onode_block = disk_getblock(disk,BR_SECTOR_T(disk,onode_br));
  onode = (struct fcfs_onode1*)onode_block->data;

  if(forknr > onode->forks.leaf.nr)
    {
      fprintf(stderr,"Invalid Fork Number for onode\n");
      abort();
    }
  if(content_length==0)
    {
      fprintf(stderr,"Must have some content to write\n");
      abort();
    }

  if((pos+content_length) > onode->forks.leaf.fork[forknr].content_length)
    {				/* Grow Onode Fork */
      if(content_length > (disk->bsize - onode->forks.leaf.fork[forknr].content_length%disk->bsize) || onode->forks.leaf.fork[forknr].content_length%disk->bsize==0)
	{
#ifdef DEBUG_ONODE_WRITE
	  fprintf(stderr,"GROWING in WRITE\n");
#endif
	  onode1_fork_grow(disk,onode_br,forknr,
			   (pos+content_length
			    - onode->forks.leaf.fork[forknr].content_length)
			   / disk->bsize
			   );
	}
      else
	fprintf(stderr,"ENOUGH SPACE IN EXISTING BLOCK\n");
    }

  done_length = pos%disk->bsize;
  do
    {
      //      fprintf(stderr,"DEBUG---- content %llu done %llu\n",content_length,done_length);

      if(content_length<= disk->bsize)
	if((disk->bsize - done_length%disk->bsize) < content_length)
	  {//fprintf(stderr,"\n\n\t\t1\n\n"); 
	    write_length = disk->bsize - done_length%disk->bsize;}
	else
	  {//fprintf(stderr,"\n\n\t\t2\n\n");
	    write_length = content_length;}
      else
	{//fprintf(stderr,"\n\n\t\t3\n\n");
	  write_length = disk->bsize - (done_length%disk->bsize);}
      
#ifdef DEBUG_ONODE_WRITE
      fprintf(stderr,"&&&&&&&&&&&WRITE LENGTH = %llu, %llu, %llu\n",write_length,pos,done_length);
#endif

      //      fprintf(stderr,"^^^^^^^^^^^GETTING BLOCK = %llu\n",done_length/disk->bsize);
      block = onode1_fork_getblock(disk,onode,forknr,(pos + (done_length - pos%disk->bsize)) / disk->bsize);
      
      memcpy(block->data+(done_length%disk->bsize),content+(done_length - pos%disk->bsize),write_length);

      onode->forks.leaf.fork[forknr].content_length+=write_length;
      content_length-=write_length;
      done_length+=write_length;

      disk_writeblock(block);
      disk_freeblock(block);
      //      fprintf(stderr,"!!!!!!!!!!!!!!LOOPING WITH %llu, %llu\n",content_length,done_length);
    } while(content_length > 0);

  //  fprintf(stderr,"\n\nDONE WRITE()\n\n");

  disk_writeblock(onode_block);
  disk_freeblock(onode_block);
}

u64 onode1_fork_read(struct fcfs_disk *disk, struct fcfs_onode1 *onode,int forknr, u64 pos, u64 content_length, void* content)
{
  struct fcfs_disk_block *block;
  u64 read_length;
  u64 done_length;
  
  if(forknr > onode->forks.leaf.nr)
    {
      fprintf(stderr,"Invalid Fork Number for onode\n");
      abort();
    }
  if(content_length==0)
    {
      fprintf(stderr,"Must have some content to read\n");
      abort();
    }

  if(pos > onode->forks.leaf.fork[forknr].content_length)
    {
      fprintf(stderr,"ERROR IN READ(): pos > content_length\n");
      return 0;
    }

  if((pos+content_length) > onode->forks.leaf.fork[forknr].content_length)
    {
      fprintf(stderr,"ERROR IN READ(): pos+content_length > content_length\n");
      content_length = onode->forks.leaf.fork[forknr].content_length - pos;
    }

  done_length = pos%disk->bsize;
  do
    {
#ifdef DEBUG_ONODE_READ
      fprintf(stderr,"DEBUG---- content %llu done %llu\n",content_length,done_length);
#endif

      if(content_length<= disk->bsize)
	if((disk->bsize - done_length%disk->bsize) < content_length)
	  {read_length = disk->bsize - done_length%disk->bsize;}
	else
	  {read_length = content_length;}
      else
	{read_length = disk->bsize - (done_length%disk->bsize);}
      
#ifdef DEBUG_ONODE_READ
      fprintf(stderr,"&&&&&&&&&&&READ LENGTH = %llu\n",read_length);

      fprintf(stderr,"^^^^^^^^^^^GETTING BLOCK = %llu\n",done_length/disk->bsize);
#endif
      block = onode1_fork_getblock(disk,onode,forknr,(pos + (done_length - pos%disk->bsize)) / disk->bsize);
      
      memcpy(content+(done_length - pos%disk->bsize),block->data+(done_length%disk->bsize),read_length);

      content_length-=read_length;
      done_length+=read_length;

      disk_freeblock(block);
#ifdef DEBUG_ONODE_READ
      fprintf(stderr,"!!!!!!!!!!!!!!LOOPING WITH %llu, %llu\n",content_length,done_length);
#endif
    } while(content_length > 0);
#ifdef DEBUG_ONODE_READ
  fprintf(stderr,"\n\nDONE READ()\n\n");
#endif
  return (done_length - pos%disk->bsize);
}


int onode1_fork_grow(struct fcfs_disk *disk, struct fcfs_block_run *onode_br, int forknr, u64 blocksnr)
{
  struct fcfs_block_run *newbr;
  struct fcfs_onode1 *onode;
  struct fcfs_disk_block *block;
  struct fcfs_disk_block *blockz;
  int leafnr;
  int i;
  
  block = disk_getblock(disk,BR_SECTOR_T(disk,onode_br));
  onode = (struct fcfs_onode1*)block->data;
  if(test_bit(FCFS_FORK_InForkData,&(onode->forks.leaf.fork[forknr].fork_flags)))
    {
      /* We're changing a InFork Fork - and we shouldn't be! */
      fprintf(stderr,"ERROR: Trying to grow InForkData fork!\n");
      abort();
    }
  else
    {
      /* We're changing a normal fork */
      if(onode->forks.leaf.fork[forknr].data.space.leaf.nr >= FSFS_ONODE1_SPACE_LEAF_MAXNR)
	{
	  fprintf(stderr,"FIXME-ERROR: Growing fork beyond leaf size\n");
	  abort();
	}
      else
	{
#ifdef ONODE_GROW_DEBUG
	  fprintf(stderr,"ONODE GROW\n");
#endif
	  leafnr = onode->forks.leaf.fork[forknr].data.space.leaf.nr;
	  //	  printf("onode_fork_grow\n");
	  if(leafnr)
	    newbr = ag_allocate_block(disk,
				      onode->forks.leaf.fork[forknr].data.space.leaf.br[leafnr-1].allocation_group,
				      onode->forks.leaf.fork[forknr].data.space.leaf.br[leafnr-1].start,
				      //+onode->forks.leaf.fork[forknr].data.space.leaf.br[leafnr-1].len-1,
				      //onode_br->start,
				      blocksnr);
	  else
	    newbr = ag_allocate_block(disk,
				      onode_br->allocation_group,
				      300,
				      blocksnr);


#ifdef ONODE_GROW_DEBUG
	  fprintf(stderr,"\n\t\t\t\t\tFORKNR INFO %u %u %u %u\n\n",
		  onode->forks.leaf.fork[forknr].data.space.leaf.br[leafnr-1].allocation_group,
		  onode->forks.leaf.fork[forknr].data.space.leaf.br[leafnr-1].start,
		  onode->forks.leaf.fork[forknr].data.space.leaf.br[leafnr-1].len,newbr->start);
#endif
	  
	  if(leafnr>0 && 
	     onode->forks.leaf.fork[forknr].data.space.leaf.br[leafnr-1].allocation_group == newbr->allocation_group
	     && onode->forks.leaf.fork[forknr].data.space.leaf.br[leafnr-1].start
	        + onode->forks.leaf.fork[forknr].data.space.leaf.br[leafnr-1].len
	     == newbr->start
	     )
	    {			/* Extend current block_run */
#ifdef ONODE_GROW_DEBUG
	      fprintf(stderr,"Extended Existing Block Run! by %u\n",newbr->len);
#endif
	      onode->forks.leaf.fork[forknr].data.space.leaf.br[leafnr-1].len+=
		newbr->len;
	    }
	  else
	    {
	      onode->forks.leaf.fork[forknr].data.space.leaf.nr++;
	      leafnr = onode->forks.leaf.fork[forknr].data.space.leaf.nr;
	      
	      onode->forks.leaf.fork[forknr].data.space.leaf.br[leafnr-1].allocation_group = newbr->allocation_group;
	      onode->forks.leaf.fork[forknr].data.space.leaf.br[leafnr-1].start = newbr->start;
	      onode->forks.leaf.fork[forknr].data.space.leaf.br[leafnr-1].len = newbr->len;
#ifdef ONODE_GROW_DEBUG
	      fprintf(stderr,"ONODE GROWN WITH BR: %u %u %u #%u\n",
		      newbr->allocation_group,
		      newbr->start,
		      newbr->len,
		      onode->forks.leaf.fork[forknr].data.space.leaf.nr
		      );
#endif
	    }
	  for(i=BR_SECTOR_T(disk,newbr);i<BR_SECTOR_T(disk,newbr)+newbr->len;i++)
	    {
	      blockz = disk_newblock(disk,i);
	      //disk_writeblock(blockz); // we don't write, as we're about
	                                 // to anyway (after return)
	      disk_freeblock(blockz);
	    }
	  free(newbr);
	}
    }
  disk_writeblock(block);
  disk_freeblock(block);
  return 1;
}
