/*
  block_dev.c
  -----------
  $Id: block_dev.c,v 1.15 2003/10/01 05:41:57 stewart Exp $
  (C)2003 Stewart Smith
  Distributed under the GNU Public License

    Some data structures have been constructed out of those
  present in the Linux Kernel (v2.5.69). They are copyright
  of their respective owners.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "block_dev.h"
#include <glib.h>



GList* lru_list;
GHashTable* blk_hash;

/* hash_bh
   -------
   Hashes a buffer head. Poor hash function, but
   okay for our purposes as we generally only
   use one block device at once in simulation.

   Used with the Glib hash.
*/
guint hash_bh(gconstpointer a)
{
  struct buffer_head *bh = (struct buffer_head*)a;
  return (bh->b_bdev->file_on_disk % 32768)+(bh->b_blocknr % 32768);
}

/* bh_equal
   --------
   true if the contents of two buffer heads are
   the same (i.e. device, block number and block size)
 */
gint bh_equal(gconstpointer a,gconstpointer b)
{
  struct buffer_head *bha,*bhb;
  bha = (struct buffer_head*)a;
  bhb = (struct buffer_head*)b;

  if(bha==NULL||bhb==NULL)
    return 0;

  return (bha->b_bdev == bhb->b_bdev
	  && bha->b_blocknr==bhb->b_blocknr
	  && bha->b_size == bhb->b_size
	  );
}

/* block_dev_init
   --------------

   Initializes the block device simulator.

   Should be called before any other function.
   On failure, aborts
 */
void block_dev_init()
{
  lru_list = NULL;
  if((blk_hash = g_hash_table_new(hash_bh,bh_equal))==NULL)
    { fprintf(stderr,"Failed to create blk_hash\n"); abort();}
}

/* block_dev_new
   -------------
   Fills out the block_device structure. doesn't allocate memory for it.

   On failure of opening teh device, aborts.
 */
int block_dev_new(struct block_device *b, const char* file, u64 block_size, u64 num_blocks)
{
  if( (b->file_on_disk=open(file,O_RDWR|O_CREAT,S_IRUSR|S_IWUSR)) < 0)
    {
      fprintf(stderr,"Unable to open device file %s\n",file);
      abort();
    }
  b->block_size = block_size;
  b->num_blocks = num_blocks;
  b->name = file;

  b->cache_hit = 0;
  b->cache_hit_clear = 0;
  b->cache_miss = 0;
  b->cache_miss_clear = 0;

  return 1;
}

/* block_dev_close
   ---------------
   Closes our simulated block device. Warns if any dirty buffers
   removes buffers from the buffer cache.
*/
int block_dev_close(struct block_device *b)
{
  struct buffer_head* bh;

  if(g_list_first(lru_list))
    {
      bh = (struct buffer_head*)(g_list_first(lru_list)->data);
      do
	{
	  if(bh->b_bdev == b)	/* Our block device */
	    {
	      if(buffer_dirty(bh))	/* Destroying without flush */
		fprintf(stderr,     /* So we warn */
			"block_dev_close: bdev has Dirty buffers (%x)\n",
			bh->b_blocknr);
	      g_hash_table_remove(blk_hash,bh);
	      lru_list = g_list_remove(lru_list,bh);
	      fprintf(stderr,"Removing block from cache 0x%08llx\n",bh->b_blocknr);
	    }
	  else
	    fprintf(stderr,"bh->b_dev = %p not %p\n",bh->b_bdev,b);
	  if(g_list_first(lru_list))
	    bh = (struct buffer_head*)g_list_first(lru_list)->data;
	  else
	    bh = NULL;
	}while(bh!=NULL);
    }
  fprintf(stderr,"Closed Block Device: %s\n",b->name);
  fprintf(stderr,"%s Stats:\n\t%lld hit\n\t%lld miss\n\n\t%lld Clear Hit\n\t%lld Clear Miss\n",b->name,b->cache_hit,b->cache_miss,b->cache_hit_clear,b->cache_miss_clear);
}

struct buffer_head *bread(struct block_device *b, sector_t block, int size)
{
  struct buffer_head *bh , *bh2;
  int i;

#ifdef DEBUG_VERBOSE
  fprintf(stderr,"Attempting to get block 0x%llx\t",block);
#endif

  if(block >= b->num_blocks)
    {
      fprintf(stderr,"block out of device range. b=%d\n",block);
      return NULL;
    }

  if((bh = (struct buffer_head*) malloc(sizeof(struct buffer_head)))==NULL)
    {
      fprintf(stderr,"Cannot allocate struct buffer_head\n");
      abort();
    }
  bh->b_blocknr = block;
  bh->b_size = size;
  bh->b_bdev = b;

  if((bh2 = g_hash_table_lookup(blk_hash,bh))!=NULL)
    {
      //  fprintf(stderr,"Found block in cache 0x%llx at 0x%x\n",block,hash_bh(bh2));
      free(bh);
      lru_list = g_list_remove(lru_list,bh2);
      lru_list = g_list_append(lru_list,bh2);
      b->cache_hit++;
      return bh2;
    }

  bh->b_state=0;
  bh->b_count=0;

  if((bh->b_data = (char*)malloc(sizeof(char)*size))==NULL)
  {
    fprintf(stderr,"Unable to Allocate data buffer\n");
    abort();
  }
  
  lseek(b->file_on_disk,b->block_size*block,SEEK_SET);
  if((i=(read(b->file_on_disk,bh->b_data,(size_t)size)!=size)))
    {
      fprintf(stderr,"An unexpected number of bytes was read: %d\n\n",i);
      abort();
    }

  lru_list = g_list_append(lru_list,bh);
  if(g_hash_table_lookup(blk_hash,bh)!=NULL && !bh_equal(g_hash_table_lookup(blk_hash,bh),bh))
    fprintf(stderr,"HASH COLLISION!!!!!\n\n");
  g_hash_table_insert(blk_hash,bh,bh);
#ifdef DEBUG_VERBOSE
  fprintf(stderr,"blocks read: %d, hash size: %d hashed as: 0x%x\n",g_list_length(lru_list),g_hash_table_size(blk_hash),hash_bh(bh));
#endif

  b->cache_miss++;

  return bh;
}

struct buffer_head *bnew(struct block_device *b, sector_t block, int size)
{
  struct buffer_head *bh , *bh2;
  int i;

#ifdef DEBUG_VERBOSE
  fprintf(stderr,"Attempting to get block 0x%llx\t",block);
#endif

  if(block >= b->num_blocks)
    {
      fprintf(stderr,"block out of device range. b=%d\n",block);
      return NULL;
    }

  if((bh = (struct buffer_head*) malloc(sizeof(struct buffer_head)))==NULL)
    {
      fprintf(stderr,"Cannot allocate struct buffer_head\n");
      abort();
    }
  bh->b_blocknr = block;
  bh->b_size = size;
  bh->b_bdev = b;

  if((bh2 = g_hash_table_lookup(blk_hash,bh))!=NULL)
    {
      //  fprintf(stderr,"Found block in cache 0x%llx at 0x%x\n",block,hash_bh(bh2));
      free(bh);
      lru_list = g_list_remove(lru_list,bh2);
      lru_list = g_list_append(lru_list,bh2);
      memset(bh2->b_data,0,size);
      b->cache_hit_clear++;
      return bh2;
    }

  bh->b_state=0;
  bh->b_count=0;

  if((bh->b_data = (char*)malloc(sizeof(char)*size))==NULL)
  {
    fprintf(stderr,"Unable to Allocate data buffer\n");
    abort();
  }
  
  memset(bh->b_data,0,size);

  lru_list = g_list_append(lru_list,bh);
  if(g_hash_table_lookup(blk_hash,bh)!=NULL && !bh_equal(g_hash_table_lookup(blk_hash,bh),bh))
    fprintf(stderr,"HASH COLLISION!!!!!\n\n");
  g_hash_table_insert(blk_hash,bh,bh);
#ifdef DEBUG_VERBOSE
  fprintf(stderr,"blocks read: %d, hash size: %d hashed as: 0x%x\n",g_list_length(lru_list),g_hash_table_size(blk_hash),hash_bh(bh));
#endif

  b->cache_miss_clear++;

  return bh;
}

void submit_bh(int rw,struct buffer_head * bh)
{
  int result;

  switch(rw)
    {
    case READ:
      break;
      
    case WRITE:
      if(!buffer_dirty(bh))
	fprintf(stderr,"Warning: Non dirty buffer being written.\n");
#ifdef DEBUG_VERBOSE
      fprintf(stderr,"---BLOCK WRITE--- %d %lld\n",bh->b_size,bh->b_blocknr);
#endif
      lseek(bh->b_bdev->file_on_disk,bh->b_size*bh->b_blocknr,SEEK_SET);
      result = write(bh->b_bdev->file_on_disk,bh->b_data,bh->b_size);
      if(result <= 0)
	{ fprintf(stderr,"Error Writing Block, result = %d\n",result);}
      clear_buffer_dirty(bh);
      fdatasync(bh->b_bdev->file_on_disk);
      break;
    default:
      fprintf(stderr,"Invalid mode to submit_bh (mode=%x)\n",rw);
      abort();
      break;
    };
}
